"""
Configuration file for the Sphinx documentation builder:

https://www.sphinx-doc.org/

"""

import sys
from importlib.metadata import version as get_version

extensions = [
    "notfound.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinxext.opengraph",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "akismet"
copyright = "Michael Foord, James Bennett, and contributors"
version = get_version("akismet")
release = version
exclude_trees = ["_build"]
pygments_style = "sphinx"
htmlhelp_basename = "akismetdoc"
latex_documents = [
    (
        "index",
        "akismet.tex",
        "akismet Documentation",
        "Michael Foord and James Bennett",
        "manual",
    ),
]
html_theme = "furo"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# Spelling check needs an additional module that is not installed by default.
# Add it only if spelling check is requested so docs can be generated without it.
if "spelling" in sys.argv:
    extensions.append("sphinxcontrib.spelling")

# Spelling language.
spelling_lang = "en_US"

# Location of word list.
spelling_word_list_filename = "spelling_wordlist.txt"

# OGP metadata configuration.
ogp_enable_meta_description = True
ogp_site_url = "https://akismet.readthedocs.io/"



"""
Common definitions used by both the sync and async Akismet implementations.

"""

# SPDX-License-Identifier: BSD-3-Clause

import enum
import os
import sys
import textwrap
import typing

import httpx

from . import _exceptions, _version

# Private constants.
# -------------------------------------------------------------------------------

_API_URL = "https://rest.akismet.com"
_API_V11 = "1.1"
_API_V12 = "1.2"
_COMMENT_CHECK = "comment-check"
_KEY_SITES = "key-sites"
_REQUEST_METHODS = typing.Literal["GET", "POST"]  # pylint: disable=invalid-name
_SUBMISSION_RESPONSE = "Thanks for making the web a better place."
_SUBMIT_HAM = "submit-ham"
_SUBMIT_SPAM = "submit-spam"
_USAGE_LIMIT = "usage-limit"
_VERIFY_KEY = "verify-key"

_KEY_ENV_VAR = "PYTHON_AKISMET_API_KEY"
_URL_ENV_VAR = "PYTHON_AKISMET_BLOG_URL"

_TIMEOUT = float(os.getenv("PYTHON_AKISMET_TIMEOUT", "1.0"))

_OPTIONAL_KEYS = [
    "blog_charset",
    "blog_lang",
    "comment_author",
    "comment_author_email",
    "comment_author_url",
    "comment_content",
    "comment_context",
    "comment_date_gmt",
    "comment_post_modified_gmt",
    "comment_type",
    "honeypot_field_name",
    "is_test",
    "permalink",
    "recheck_reason",
    "referrer",
    "user_agent",
    "user_role",
]


# Public constants.
# -------------------------------------------------------------------------------

USER_AGENT = (
    f"akismet.py/{_version.LIBRARY_VERSION} | Python/"
    f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
)


# Public classes.
# -------------------------------------------------------------------------------


class CheckResponse(enum.IntEnum):
    """
    Possible response values from an Akismet content check, including the
    possibility of the "discard" response, modeled as an :class:`enum.IntEnum`.

    """

    HAM = 0
    SPAM = 1
    DISCARD = 2


class Config(typing.NamedTuple):
    """
    A :func:`~collections.namedtuple` representing Akismet configuration, consisting
    of a key and a URL.

    You only need to use this if you're manually configuring an Akismet API client
    rather than letting the configuration be read automatically from environment
    variables.

    """

    key: str
    url: str


# Private helper functions.
# -------------------------------------------------------------------------------


def _configuration_error(config: Config) -> typing.NoReturn:
    """
    Raise an appropriate exception for invalid configuration.

    """
    raise _exceptions.APIKeyError(
        textwrap.dedent(
            f"""
            Akismet API key and/or blog URL were invalid.

            Found API key: {config.key}
            Found URL: {config.url}
            """
        )
    )


def _get_async_http_client() -> httpx.AsyncClient:
    """
    Return an asynchronous HTTP client for interacting with the Akismet API.

    """
    return httpx.AsyncClient(headers={"User-Agent": USER_AGENT}, timeout=_TIMEOUT)


def _get_sync_http_client() -> httpx.Client:
    """
    Return a synchronous HTTP client for interacting with the Akismet API.

    """
    return httpx.Client(headers={"User-Agent": USER_AGENT}, timeout=_TIMEOUT)


def _protocol_error(operation: str, response: httpx.Response) -> typing.NoReturn:
    """
    Raise an appropriate exception for unexpected API responses.

    """
    raise _exceptions.ProtocolError(
        textwrap.dedent(
            f"""
        Received unexpected or non-standard response from Akismet API.

        API operation was: {operation}
        API response received was: {response.text}
        Debug header value was: {response.headers.get('X-akismet-debug-help', None)}
        """
        )
    )


def _try_discover_config() -> Config:
    """
    Attempt to discover and return an Akismet configuration from the environment.

    :raises akismet.ConfigurationError: When either or both of the API key and
       URL are missing, or if the URL does not begin with ``"http://"`` or
       ``https://``.

    """
    key = os.getenv(_KEY_ENV_VAR, None)
    url = os.getenv(_URL_ENV_VAR, None)
    if not all([key, url]):
        raise _exceptions.ConfigurationError(
            textwrap.dedent(
                f"""
        Could not find full Akismet configuration.

        Found API key: {key}
        Found blog URL: {url}
        """
            )
        )
    if not url.startswith(("http://", "https://")):
        raise _exceptions.ConfigurationError(
            textwrap.dedent(
                f"""
            Invalid Akismet site URL specified: {url}

            Akismet requires the full URL including the leading
            'http://' or 'https://'.
            """
            )
        )
    return Config(key=key, url=url)



"""
Asynchronous Akismet API client implementation.

"""

# SPDX-License-Identifier: BSD-3-Clause

from types import TracebackType
from typing import TYPE_CHECKING, Literal, Optional, Type, Union

import httpx

from . import _common, _exceptions

if TYPE_CHECKING:  # pragma: no cover
    import akismet


class AsyncClient:
    """
    Asynchronous Akismet API client.

    All methods of the Akismet 1.1 and 1.2 web APIs are implemented here:

    * :meth:`comment_check`
    * :meth:`key_sites`
    * :meth:`submit_ham`
    * :meth:`submit_spam`
    * :meth:`usage_limit`
    * :meth:`verify_key`

    Use of this client requires an Akismet API key; see <https://akismet.com> for
    instructions on how to obtain one. Once you have an Akismet API key and
    corresponding registered site URL to use with it, you can create an API client in
    either of two ways.

    **Recommended for most uses:** Place your Akismet API key and site URL in the
    environment variables ``PYTHON_AKISMET_API_KEY`` and ``PYTHON_AKISMET_BLOG_URL``,
    and then use the :meth:`validated_client` constructor:

    .. code-block:: python

       import akismet
       akismet_client = await akismet.AsyncClient.validated_client()

    This will automatically read the API key and site URL from the environment
    variables, instantiate a client, and use its :meth:`verify_key` method to ensure the
    key and URL are valid before returning the client instance to you. See :ref:`the FAQ
    <alt-constructor>` for the technical reasons why the default constructor does not
    have this behavior.

    If you don't want to or can't use the environment variables to configure Akismet,
    you can also explicitly configure by creating a :class:`~akismet.Config` instance
    with your API key and site URL, and passing it as the constructor argument
    ``config``:

    .. code-block:: python

       import akismet
       config = akismet.Config(key=your_api_key, url=your_site_url)
       akismet_client = await akismet.AsyncClient.validated_client(config=config)

    If you rely on environment variable configuration and the complete configuration
    cannot be found in the environment variables, :meth:`validated_client` will raise
    :exc:`~akismet.ConfigurationError`. If the API key and URL you supply are invalid
    according to :meth:`verify_key` -- regardless of whether you provided them via
    environment variables or an explicit :class:`~akismet.Config` --
    :meth:`validated_client` will raise :exc:`~akismet.APIKeyError`.

    If you want to modify the HTTP request behavior -- for example, to support a
    required HTTP proxy -- you can construct a custom ``httpx.AsyncClient`` and pass it
    as the keyword argument ``http_client`` to either :meth:`validated_client` or the
    default constructor. See :data:`akismet.USER_AGENT` for the default user-agent
    string used by the Akismet API clients, and <https://www.python-httpx.org> for the
    full documentation of the HTTPX module.

    Note that if you only want to set a custom request timeout threshold (the default is
    1 second), you can specify it by setting the environment variable
    ``PYTHON_AKISMET_TIMEOUT`` to a value that can be parsed into a :class:`float` or
    :class:`int` and represents the desired timeout in seconds.

    You can also use this class as a context manager; when doing so, you do *not* need
    to use the :meth:`validated_client` constructor, as the context manager can perform
    the validation for you when entering the ``with`` block:

    .. code-block:: python

       import akismet

       async with akismet.AsyncClient() as akismet_client:
           # Use the client here. It will be automatically cleaned up when the "with"
           # block exits.

    **Unusual/advanced use cases:** Invoke the default constructor. It accepts the same
    set of arguments as the :meth:`validated_client` constructor, and its behavior is
    identical *except* for the fact that it will not automatically validate your
    configuration, so you must remember to do so manually. You should only invoke the
    default constructor if you are absolutely certain that you need to avoid the
    automatic validation performed by :meth:`validated_client`.

    .. warning:: **Consequences of invalid configurationn**

       If you construct an Akismet API client through the default constructor and
       provide an invalid key or URL, all operations of the Akismet web service, other
       than key verification, will reply with an invalid-key message. This will cause
       all client methods other than :meth:`verify_key` to raise
       :exc:`~akismet.APIKeyError`. To avoid this situation, it is strongly recommended
       that you call :meth:`verify_key` to validate your configuration prior to calling
       any other methods, at which point you likely should be using
       :meth:`validated_client` anyway.

    :param config: An optional Akismet :class:`~akismet.Config`, consisting of an API
       key and site URL.

    :param http_client: An optional ``httpx`` async HTTP client instance to
       use. Generally you should only pass this in if you need significantly customized
       HTTP-client behavior, and if you do pass this argument you are responsible for
       setting an appropriate ``User-Agent`` (see :data:`~akismet.USER_AGENT`), timeout,
       and other configuration values. If all you want is to change the default timeout
       (1 second), store the desired timeout, in seconds, as a floating-point or integer
       value in the environment variable ``PYTHON_AKISMET_TIMEOUT``.

    """

    _http_client: httpx.AsyncClient
    _config: "akismet.Config"

    # Constructors.
    # ----------------------------------------------------------------------------

    def __init__(
        self,
        config: Optional["akismet.Config"] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ) -> None:
        """
        Default constructor.

        You will almost always want to use :meth:`validated_client` instead.

        """
        self._config = config if config is not None else _common._try_discover_config()
        self._http_client = http_client or _common._get_async_http_client()

    @classmethod
    async def validated_client(
        cls,
        config: Optional["akismet.Config"] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ) -> "AsyncClient":
        """
        Constructor of :class:`AsyncClient`.

        This is usually preferred over the default ``AsyncClient()`` constructor,
        because this constructor will validate the Akismet configuration (API key and
        URL) prior to returning the client instance.

        :param config: An optional explicit Akismet :class:`~akismet.Config`, consisting
           of an API key and site URL; if not passed, the configuration will be read
           from the environment variables ``PYTHON_AKISMET_API_KEY`` and
           ``PYTHON_AKISMET_BLOG_URL``.

        :param http_client: An optional ``httpx`` async HTTP client instance to
           use. Generally you should only pass this in if you need significantly
           customized HTTP-client behavior, and if you do pass this argument you are
           responsible for setting an appropriate ``User-Agent`` (see
           :data:`~akismet.USER_AGENT`), timeout, and other configuration values. If all
           you want is to change the default timeout (1 second), store the desired
           timeout, in seconds, as a floating-point or integer value in the environment
           variable ``PYTHON_AKISMET_TIMEOUT``.

        :raises akismet.APIKeyError: When the discovered Akismet configuration is
           invalid according to :meth:`verify_key`.

        :raises akismet.ConfigurationError: When the Akismet configuration is partially
           or completely missing, or when the supplied site URL is in the wrong format
           (does not begin with ``http://`` or ``https://``).

        """
        # While the synchronous version of the client could perform the config discovery
        # and validation in __init__(), here we cannot because this client's
        # verify_key() method is async, and its underlying HTTP client is async. So
        # calling into them would require making __init__ into an async method, and
        # Python does not currently allow __init__() to be usefully async. But a
        # classmethod *can* be async, so we define and encourage the use of an
        # alternative constructor in order to achieve API consistency.
        instance = cls(config=config, http_client=http_client)
        if not await instance.verify_key():
            _common._configuration_error(instance._config)
        return instance

    # Async context-manager protocol.
    # ----------------------------------------------------------------------------

    async def __aenter__(self) -> "AsyncClient":
        """
        Entry method of the async context manager.

        """
        if not await self.verify_key():
            _common._configuration_error(self._config)
        return self

    async def __aexit__(
        self, exc_type: Type[BaseException], exc: BaseException, tb: TracebackType
    ):
        """
        Exit method of the async context manager.

        """
        await self._http_client.aclose()

    # Internal/helper methods.
    # ----------------------------------------------------------------------------

    async def _request(
        self,
        method: _common._REQUEST_METHODS,
        version: str,
        endpoint: str,
        data: dict,
    ) -> httpx.Response:
        """
        Make a request to the Akismet API and return the response.

        :param method: The HTTP request method to use.

        :param version: The Akismet API version to use.

        :param endpoint: The Akismet API endpoint to post to.

        :param data: The data to send in the request.

        :raises akismet.RequestError: When an error occurs connecting to Akismet, or
           when Akiset returns a non-success status code.

        """
        method = method.upper()
        if method not in ("GET", "POST"):
            raise _exceptions.AkismetError(
                f"Unrecognized request method attempted: {method}."
            )
        handler = getattr(self._http_client, method.lower())
        request_kwarg = "data" if method == "POST" else "params"
        try:
            response = await handler(
                f"{_common._API_URL}/{version}/{endpoint}", **{request_kwarg: data}
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise _exceptions.RequestError(
                f"Akismet responded with error status: {exc.response.status_code}"
            ) from exc
        except httpx.TimeoutException as exc:
            raise _exceptions.RequestError("Akismet timed out.") from exc
        except httpx.RequestError as exc:
            raise _exceptions.RequestError("Error making request to Akismet.") from exc
        except Exception as exc:
            raise _exceptions.RequestError("Error making request to Akismet.") from exc
        # Since it's possible to construct a client without performing up-front API key
        # validation, we have to watch out here for the possibility that we're making
        # requests with an invalid key, and raise the appropriate exception.
        if endpoint != _common._VERIFY_KEY and response.text == "invalid":
            raise _exceptions.APIKeyError(
                "Akismet API key and/or site URL are invalid."
            )
        return response

    async def _get_request(
        self, version: str, endpoint: str, params: dict
    ) -> httpx.Response:
        """
        Make a GET request to the Akismet API and return the response.

        This method is used by most HTTP GET API calls.

        :param version: The Akismet API version to use.

        :param endpoint: The Akismet API endpoint to post to.

        :param params: The querystring parameters to include in the request.

        :raises akismet.APIKeyError: When the configured API key and/or site URL are
           invalid.

        """
        return await self._request("GET", version, endpoint, params)

    async def _post_request(
        self, version: str, endpoint: str, user_ip: str, **kwargs: str
    ) -> httpx.Response:
        """
        Make a POST request to the Akismet API and return the response.

        This method is used by most HTTP POST API calls except key verification.

        :param version: The Akismet API version to use.

        :param endpoint: The Akismet API endpoint to post to.

        :param user_ip: The IP address of the user who submitted the content.

        :raises akismet.APIKeyError: When the configured API key and/or site URL are
           invalid.

        :raises akismet.UnknownArgumentError: When one or more unexpected optional
           argument names are supplied. See `the Akismet documentation
           <https://akismet.com/developers/comment-check/>`_ for details of supported
           optional argument names.

        """
        unknown_args = [k for k in kwargs if k not in _common._OPTIONAL_KEYS]
        if unknown_args:
            raise _exceptions.UnknownArgumentError(
                f"Received unknown argument(s) for Akismet operation {endpoint}: "
                f"{', '.join(unknown_args)}"
            )
        data = {
            "api_key": self._config.key,
            "blog": self._config.url,
            "user_ip": user_ip,
            **kwargs,
        }
        return await self._request("POST", version, endpoint, data)

    async def _submit(self, endpoint: str, user_ip: str, **kwargs: str) -> bool:
        """
        Submit ham or spam to the Akismet API.

        :param endpoint: The endpoint (either ``""submit-ham""`` or ``""submit-spam""``)
           to send the content to.

        :param user_ip: The IP address of the user who submitted the content.

        :raises akismet.ProtocolError: When an unexpected/invalid response type is
           received from the Akismet API.

        """
        response = await self._post_request(
            _common._API_V11, endpoint, user_ip=user_ip, **kwargs
        )
        if response.text == _common._SUBMISSION_RESPONSE:
            return True
        _common._protocol_error(endpoint, response)

    # Public methods corresponding to the methods of the Akismet API.
    # ----------------------------------------------------------------------------

    async def comment_check(
        self, user_ip: str, **kwargs: str
    ) -> "akismet.CheckResponse":
        """
        Check a piece of user-submitted content to determine whether it is spam.

        The IP address of the user posting the content is required. All `other
        comment-check arguments documented by Akismet
        <https://akismet.com/developers/comment-check/>`_ are also optionally accepted.

        It is recommended that you supply at least the following optional arguments:
        ``comment_content``; ``comment_type``; and ``comment_author`` and/or
        ``comment_author_email``.

        The return value is an :class:`int` from the :class:`~akismet.CheckResponse`
        enum, which can be used as a truthy value (``0``/:data:`False` if the content is
        not classified as spam, ``1``/:data:`True` if it is classified as spam). But
        making use of the full set of enum values allows detecting the presence of `the
        "discard" value <https://akismet.com/blog/theres-a-ninja-in-your-akismet/>`_ in
        the ``X-akismet-pro-tip`` header to indicate "blatant" spam.

        :param user_ip: The IP address of the user who submitted the content.

        :param str comment_content: (optional, recommended) The content the user
           submitted.

        :param str comment_type: (optional, recommended) The type of content, with
           common values being ``"comment"``, ``"forum-post"``, ``"contact-form"``, and
           ``"signup"``. See the Akismet service documentation for a full list of
           common/recommended types.

        :param str comment_author: (optional, recommended) The name (such as username)
           of the content's submitter.

        :param str comment_author_email: (optional, recommended) The email address of
           the content's submitter.

        :param int is_test: (optional) Set to ``1`` if you are making requests for
          testing purposes; this tells Akismet not to incorporate the request into its
          training corpus or allow it to affect future responses.

        :raises akismet.ProtocolError: When an unexpected/invalid response type is
           received from the Akismet API.

        """
        response = await self._post_request(
            _common._API_V11, _common._COMMENT_CHECK, user_ip=user_ip, **kwargs
        )
        if response.text == "true":
            if response.headers.get("X-akismet-pro-tip", "") == "discard":
                return _common.CheckResponse.DISCARD
            return _common.CheckResponse.SPAM
        if response.text == "false":
            return _common.CheckResponse.HAM
        _common._protocol_error(_common._COMMENT_CHECK, response)

    async def submit_ham(self, user_ip: str, **kwargs: str) -> bool:
        """
        Inform Akismet that a piece of user-submitted comment is not spam.

        The IP address of the user posting the content is required. All `other
        submit-ham arguments documented by Akismet
        <https://akismet.com/developers/submit-ham/>`_ are also optionally accepted.

        It is recommended that you supply at least the following optional arguments:
        ``comment_content``; ``comment_type``; and ``comment_author`` and/or
        ``comment_author_email``.

        Will return :data:`True` on success (the only expected response).

        :param user_ip: The IP address of the user who submitted the content.

        :param str comment_content: (optional, recommended) The content the user
           submitted.

        :param str comment_type: (optional, recommended) The type of content, with
           common values being ``"comment"``, ``"forum-post"``, ``"contact-form"``, and
           ``"signup"``. See the Akismet service documentation for a full list of
           common/recommended types.

        :param str comment_author: (optional, recommended) The name (such as username)
           of the content's submitter.

        :param str comment_author_email: (optional, recommended) The email address of
           the content's submitter.

        :param int is_test: (optional) Set to ``1`` if you are making requests for
          testing purposes; this tells Akismet not to incorporate the request into its
          training corpus or allow it to affect future responses.

        :raises akismet.ProtocolError: When an unexpected/invalid response type is
           received from the Akismet API.

        """
        return await self._submit(_common._SUBMIT_HAM, user_ip, **kwargs)

    async def submit_spam(self, user_ip: str, **kwargs: str) -> bool:
        """
        Inform Akismet that a piece of user-submitted comment is spam.

        The IP address of the user posting the content is required. All `other
        submit-spam arguments documented by Akismet
        <https://akismet.com/developers/submit-spam/>`_ are also optionally accepted.

        It is recommended that you supply at least the following optional arguments:
        ``comment_content``; ``comment_type``; and ``comment_author`` and/or
        ``comment_author_email``.

        Will return :data:`True` on success (the only expected response).

        :param user_ip: The IP address of the user who submitted the content.

        :param str comment_content: (optional, recommended) The content the user
           submitted.

        :param str comment_type: (optional, recommended) The type of content, with
           common values being ``"comment"``, ``"forum-post"``, ``"contact-form"``, and
           ``"signup"``. See the Akismet service documentation for a full list of
           common/recommended types.

        :param str comment_author: (optional, recommended) The name (such as username)
           of the content's submitter.

        :param str comment_author_email: (optional, recommended) The email address of
           the content's submitter.

        :param int is_test: (optional) Set to ``1`` if you are making requests for
          testing purposes; this tells Akismet not to incorporate the request into its
          training corpus or allow it to affect future responses.

        :raises akismet.ProtocolError: When an unexpected/invalid response type is
           received from the Akismet API.

        """
        return await self._submit(_common._SUBMIT_SPAM, user_ip, **kwargs)

    async def key_sites(  # pylint: disable=too-many-arguments
        self,
        month: Optional[str] = None,
        url_filter: Optional[str] = None,
        result_format: Literal["csv", "json"] = "json",
        order: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[dict, str]:
        """
        Return Akismet API usage statistics keyed by site.

        All arguments are optional, and the Akismet API will set them to default values
        if not supplied.

        See `the Akismet key-sites documentation
        <https://akismet.com/developers/key-sites-activity/>`_ for examples of the
        response data from this method.

        :param month: The month, in ``"YYYY-MM"`` format, to retrieve statistics for. If
           not supplied, defaults to the current month.

        :param url_filter: A full or partial site URL to filter results by. If not
           supplied, results for all sites under the current API key will be returned.

        :param result_format: The format in which to return results. Supported options
           are ``"json"`` and ``"csv"``. Defaults to ``"json"`` if not supplied.

        :param order: For CSV-formatted results, the column by which the results should
           be sorted.

        :param limit: The maximum number of results to return. If not supplied, defaults
           to 500.

        :param offset: The offset from which to begin result reporting. If not supplied,
           defaults to 0.

        """
        params = {}
        for argument, value in (
            ("month", month),
            ("filter", url_filter),
            ("format", result_format),
            ("order", order),
            ("limit", limit),
            ("offset", offset),
        ):
            if value is not None:
                params[argument] = value
        response = await self._get_request(_common._API_V12, _common._KEY_SITES, params)
        if result_format == "csv":
            return response.text
        return response.json()

    async def usage_limit(self) -> dict:
        """
        Return Akismet API usage statistics for the current month.

        See `the Akismet usage-limit documentation
        <https://akismet.com/developers/usage-limit/>`_ for examples of the response
        data from this method.

        """
        response = await self._get_request(
            _common._API_V12, _common._USAGE_LIMIT, params={"api_key": self._config.key}
        )
        return response.json()

    async def verify_key(
        self, key: Optional[str] = None, url: Optional[str] = None
    ) -> bool:
        """
        Verify an Akismet API key and URL.

        Return :data:`True` if the key and URL are valid, :data:`False` otherwise.

        In general, you should not need to explicitly call this method. The
        :meth:`validated_client` constructor will ensure this method is called during
        client construction, after which the now-verified key/URL can be trusted. If
        neither ``key`` nor ``url`` are provided, the key and URL currently in use by
        this client will be checked.

        :param key: The API key to check.

        :param url: The URL to check.

        :raises akismet.ProtocolError: When an unexpected/invalid response type is
           received from the Akismet API.

        """
        if not all([key, url]):
            key, url = self._config
        response = await self._request(
            "POST", _common._API_V11, _common._VERIFY_KEY, {"key": key, "blog": url}
        )
        if response.text == "valid":
            return True
        if response.text == "invalid":
            return False
        _common._protocol_error(_common._VERIFY_KEY, response)



"""
Exceptions used by the Akismet clients.

"""

# SPDX-License-Identifier: BSD-3-Clause


class AkismetError(Exception):
    """
    Base exception class for Akismet errors.

    """


class UnknownArgumentError(TypeError, AkismetError):
    """
    Raised when an unknown argument was used as part of an API request.

    """


class RequestError(AkismetError):
    """
    Raised when an unexpected error occurred when making a request to Akismet.

    This is almost always going to be a `chained exception
    <https://peps.python.org/pep-3134/>`_ wrapping an underlying error from the HTTP
    client, so inspecting the exception chain may yield useful debugging information.

    """


class ProtocolError(AkismetError):
    """
    Raised when an unexpected or non-standard response was received from Akismet.

    """


class ConfigurationError(AkismetError):
    """
    Raised when an Akismet configuration error is detected (config missing or
    invalid).

    """


class APIKeyError(ConfigurationError):
    """
    Raised when the supplied Akismet API key/URL are invalid according to the
    verify-key operation.

    This is a subclass of :exc:`~akismet.ConfigurationError`. If you want to detect the
    specific case of an invalid key/URL, catch this exception, while if you just want to
    catch all configuration-related errors (including missing configuration), catch
    ``ConfigurationError`` instead.

    """



"""
Tests for the synchronous Akismet API client.

"""

# SPDX-License-Identifier: BSD-3-Clause

import csv
import os
import typing
from http import HTTPStatus
from unittest import mock

import httpx

import akismet
from akismet import _common, _test_clients

from . import base


def make_fixed_response_sync_client(
    response_text: str = "true",
    status_code: HTTPStatus = HTTPStatus.OK,
    response_json: typing.Optional[dict] = None,
) -> httpx.Client:
    """
    Return a synchronous HTTP client that produces a fixed repsonse, for use in
    testing.

    """
    return httpx.Client(
        transport=base.make_fixed_response_transport(
            response_text, status_code, response_json
        )
    )


def make_exception_sync_client(
    exception_class: Exception, message: str = "Error!"
) -> httpx.Client:
    """
    Return a synchronous HTTP client that raises the given exception/message.

    """
    return mock.Mock(
        spec_set=httpx.Client,
        get=mock.Mock(side_effect=exception_class(message)),
        post=mock.Mock(side_effect=exception_class(message)),
    )


class ValidConfig(akismet.TestSyncClient):
    """
    Test client with valid config.

    """

    verify_key_response = True


class InvalidConfig(akismet.TestSyncClient):
    """
    Test client with invalid config.

    """

    verify_key_response = False


class AlwaysBlatantSpam(akismet.TestSyncClient):
    """
    Test client which marks all content as "blatant" spam.

    """

    comment_check_response = akismet.CheckResponse.DISCARD


class AlwaysSpam(akismet.TestSyncClient):
    """
    Test client which marks all content as spam.

    """

    comment_check_response = akismet.CheckResponse.SPAM


class NeverSpam(akismet.TestSyncClient):
    """
    Test client which marks all content as not spam.

    """

    comment_check_response = akismet.CheckResponse.HAM


class SyncAkismetConstructorTests(base.AkismetTests):
    """
    Test the constructors of the synchronous Akismet API client.

    """

    def test_construct_config_explicit(self):
        """
        Passing explicit config to the default constructor uses that config.

        """
        config = akismet.Config(key="other-invalid-test-key", url=self.config.url)
        with ValidConfig(config=config) as client:
            assert client._config == config

    def test_construct_config_alternate_constructor_explicit(self):
        """
        Passing explicit config to the alternate constructor uses that config.

        """
        config = akismet.Config(key="other-invalid-test-key", url=self.config.url)
        client = ValidConfig.validated_client(config=config)
        assert client._config == config

    def test_construct_config_from_env(self):
        """
        Instantiating via the default constructor, without passing explicit config,
        reads the config from the environment.

        """
        config = akismet.Config(key=self.api_key, url=self.site_url)
        with ValidConfig() as client:
            assert client._config == config

    def test_construct_alternate_constructor_config_from_env(self):
        """
        Instantiating via the alternate constructor, without passing explicit
        config, reads the config from the environment.

        """
        config = akismet.Config(key=self.api_key, url=self.site_url)
        client = ValidConfig.validated_client()
        assert client._config == config

    def test_construct_config_valid(self):
        """
        With a valid configuration, constructing a client succeeds.

        """
        ValidConfig.validated_client()

    def test_construct_config_invalid_key(self):
        """
        With an invalid API key, constructing a client raises an APIKeyError.

        """
        with self.assertRaises(akismet.APIKeyError):
            InvalidConfig.validated_client()

    def test_construct_config_valid_context_manager(self):
        """
        With a valid configuration, constructing a client as a context manager succeeds.

        """
        with ValidConfig():
            pass

    def test_construct_config_invalid_key_context_manager(self):
        """
        With an invalid API key, constructing a client as a context manager raises
        an APIKeyError.

        """
        with self.assertRaises(akismet.APIKeyError):
            with InvalidConfig():
                pass

    def test_construct_config_valid_explicit(self):
        """
        With an explicit valid configuration, constructing a client succeeds.

        """
        ValidConfig.validated_client(config=self.config)

    def test_construct_config_invalid_key_explicit(self):
        """
        With an explicit invalid API key, constructing a client raises an APIKeyError.

        """
        with self.assertRaises(akismet.APIKeyError):
            InvalidConfig.validated_client(config=self.config)

    def test_construct_config_bad_url(self):
        """
        With an invalid URL, constructing a client raises a ConfigurationError.

        """
        try:
            os.environ[_common._URL_ENV_VAR] = "ftp://example.com"
            with self.assertRaises(akismet.ConfigurationError):
                akismet.SyncClient.validated_client()
        finally:
            os.environ[_common._URL_ENV_VAR] = self.site_url

    def test_construct_config_missing_key(self):
        """
        Without an API key present, constructing a client raises a
        ConfigurationError.

        """
        try:
            if _common._KEY_ENV_VAR in os.environ:
                del os.environ[_common._KEY_ENV_VAR]
            with self.assertRaises(akismet.ConfigurationError):
                ValidConfig.validated_client()
        finally:
            os.environ[_common._KEY_ENV_VAR] = self.api_key

    def test_construct_config_missing_url(self):
        """
        Without a registered site URL present, constructing a client raises a
        ConfigurationError.

        """
        try:
            if _common._URL_ENV_VAR in os.environ:
                del os.environ[_common._URL_ENV_VAR]
            with self.assertRaises(akismet.ConfigurationError):
                ValidConfig.validated_client()
        finally:
            os.environ[_common._URL_ENV_VAR] = self.site_url

    def test_construct_config_missing_all(self):
        """
        Without any config present, constructing a client raises a
        ConfigurationError.

        """
        try:
            if _common._KEY_ENV_VAR in os.environ:
                del os.environ[_common._KEY_ENV_VAR]
            if _common._URL_ENV_VAR in os.environ:
                del os.environ[_common._URL_ENV_VAR]
            with self.assertRaises(akismet.ConfigurationError):
                ValidConfig.validated_client()
        finally:
            os.environ[_common._KEY_ENV_VAR] = self.api_key
            os.environ[_common._URL_ENV_VAR] = self.site_url

    def test_construct_default_client(self):
        """
        Constructing a client without an explicit HTTP client uses the default HTTP
        client.

        """
        client = akismet.SyncClient()
        http_client = client._http_client
        assert "user-agent" in http_client.headers
        assert http_client.headers["user-agent"] == _common.USER_AGENT


class SyncAkismetAPITests(base.AkismetTests):
    """
    Test the API behavior of the synchronous Akismet API client.

    """

    def test_unsupported_request_method(self):
        """
        Attempting to make a request with an unsupported method raises
        ``AkismetError``.

        """
        client = ValidConfig()
        # The tested set of methods here are all the methods that are in Python 3.11's
        # http.HTTPMethod enum but not supported for Akismet requests.
        for bad_method in (
            "CONNECT",
            "DELETE",
            "HEAD",
            "OPTIONS",
            "PATCH",
            "PUT",
            "TRACE",
        ):
            with self.subTest(method=bad_method):
                with self.assertRaises(akismet.AkismetError):
                    client._request(
                        bad_method,
                        _common._API_V11,
                        _common._COMMENT_CHECK,
                        {"api_key", client._config.key},
                    )

    def test_verify_key_valid(self):
        """
        ``verify_key()`` returns True when the config is valid.

        """
        with ValidConfig() as client:
            assert client.verify_key()

    def test_verify_key_invalid(self):
        """
        ``verify_key()`` returns False when the config is invalid.

        """
        client = InvalidConfig()
        assert not client.verify_key()

    def test_verify_key_valid_explicit(self):
        """
        ``verify_key()`` returns True when the config is valid and explicitly passed
        in.

        """
        with ValidConfig() as client:
            assert client.verify_key(key=self.api_key, url=self.site_url)

    def test_verify_key_invalid_explicit(self):
        """
        ``verify_key()`` returns False when the config is invalid and explicitly
        passed in.

        """
        client = InvalidConfig()
        assert not client.verify_key(key=self.api_key, url=self.site_url)

    def test_request_with_invalid_key(self):
        """
        The request methods other than ``verify_key()`` raise
        ``akismet.APIKeyError`` if called with an invalid API key/URL.

        """
        client = akismet.SyncClient(
            http_client=make_fixed_response_sync_client(response_text="invalid"),
        )
        for method in ("comment_check", "submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.APIKeyError):
                    getattr(client, method)(**self.common_kwargs)
        for method in ("key_sites", "usage_limit"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.APIKeyError):
                    getattr(client, method)()

    def test_comment_check_spam(self):
        """
        ``comment_check()`` returns the SPAM value when Akismet declares the content
        to be spam.

        """
        with AlwaysSpam() as client:
            assert (
                client.comment_check(comment_content="test", **self.common_kwargs)
                == akismet.CheckResponse.SPAM
            )

    def test_comment_check_spam_discard(self):
        """
        ``comment_check()`` returns the DISCARD value when Akismet declares the content
        to be spam and sends the "discard"" header value.

        """
        with AlwaysBlatantSpam() as client:
            assert (
                client.comment_check(comment_content="test", **self.common_kwargs)
                == akismet.CheckResponse.DISCARD
            )

    def test_comment_check_ham(self):
        """
        ``comment_check()`` returns the HAM value when Akismet declares the content
        to be ham.

        """
        with NeverSpam() as client:
            assert (
                client.comment_check(comment_content="test", **self.common_kwargs)
                == akismet.CheckResponse.HAM
            )

    def test_submit_ham(self):
        """
        ``submit_ham()`` returns True when Akismet accepts the submission.

        """
        with ValidConfig() as client:
            assert client.submit_ham(**self.common_kwargs)

    def test_submit_spam(self):
        """
        ``submit_spam()`` returns True when Akismet accepts the submission.

        """
        with ValidConfig() as client:
            assert client.submit_spam(**self.common_kwargs)

    def test_key_sites_json(self):
        """
        ``key_sites()`` returns key usage information in JSON format by default.

        """
        with ValidConfig() as client:
            response_json = client.key_sites()
        for key in ["2022-09", "limit", "offset", "total"]:
            assert key in response_json
        sites = response_json["2022-09"]
        for site in sites:
            for key in [
                "site",
                "api_calls",
                "spam",
                "ham",
                "missed_spam",
                "false_positives",
                "is_revoked",
            ]:
                assert key in site

    def test_key_sites_csv(self):
        """
        ``key_sites()`` returns key usage information in CSV format when requested.

        """
        with ValidConfig() as client:
            first, *rest = (client.key_sites(result_format="csv")).splitlines()
        assert first.startswith("Active sites for")
        reader = csv.DictReader(rest)
        row = next(reader)
        assert set(row.keys()) == {
            "Site",
            "Total API Calls",
            "Spam",
            "Ham",
            "Missed Spam",
            "False Positives",
            "Is Revoked",
        }

    def test_usage_limit(self):
        """
        ``usage_limit()`` returns the API usage statistics in JSON format.

        """
        with ValidConfig() as client:
            response_json = client.usage_limit()
        assert set(response_json.keys()) == {
            "limit",
            "usage",
            "percentage",
            "throttled",
        }


class SyncAkismetErrorTests(base.AkismetTests):
    """
    Test the error behavior of the synchronous Akismet API client.

    """

    def test_error_status(self):
        """
        RequestError is raised when a POST request to Akismet responds with an HTTP
        status code indicating an error.

        """
        codes = [code for code in HTTPStatus if 400 <= code <= 599]
        for code in codes:
            client = akismet.SyncClient(
                http_client=make_fixed_response_sync_client(status_code=code),
            )
            with self.subTest(method="verify_key"):
                with self.assertRaises(akismet.RequestError):
                    client.verify_key()
            for method in ("comment_check", "submit_ham", "submit_spam"):
                with self.subTest(method=method):
                    with self.assertRaises(akismet.RequestError):
                        getattr(client, method)(**self.common_kwargs)
            for method in ("key_sites", "usage_limit"):
                with self.subTest(method=method):
                    with self.assertRaises(akismet.RequestError):
                        getattr(client, method)()

    def test_error_timeout(self):
        """
        RequestError is raised when the request to Akismet times out.

        """

        client = akismet.SyncClient(
            http_client=make_exception_sync_client(
                httpx.TimeoutException, "Timed out."
            ),
        )
        with self.subTest(method="verify_key"):
            with self.assertRaises(akismet.RequestError):
                client.verify_key()
        for method in ("comment_check", "submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.RequestError):
                    getattr(client, method)(**self.common_kwargs)
        for method in ("key_sites", "usage_limit"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.RequestError):
                    getattr(client, method)()

    def test_error_other_httpx(self):
        """
        RequestError is raised when a generic ``httpx`` request error occurs.

        """
        client = akismet.SyncClient(
            http_client=make_exception_sync_client(httpx.RequestError),
        )
        with self.subTest(method="verify_key"):
            with self.assertRaises(akismet.RequestError):
                client.verify_key()
        for method in ("comment_check", "submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.RequestError):
                    getattr(client, method)(**self.common_kwargs)
        for method in ("key_sites", "usage_limit"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.RequestError):
                    getattr(client, method)()

    def test_error_other(self):
        """
        RequestError is raised when any other (non-``httpx``) exception occurs during
        the request.

        """
        client = akismet.SyncClient(
            http_client=make_exception_sync_client(TypeError),
        )
        with self.subTest(method="verify_key"):
            with self.assertRaises(akismet.RequestError):
                client.verify_key()
        for method in ("comment_check", "submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.RequestError):
                    getattr(client, method)(**self.common_kwargs)
        for method in ("key_sites", "usage_limit"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.RequestError):
                    getattr(client, method)()

    def test_unknown_argument(self):
        """
        UnknownArgumentError is raised when an argument outside the supported set is
        passed to one of the POST request methods.

        """
        client = akismet.SyncClient(
            http_client=_test_clients._make_test_async_http_client()
        )
        for method in ("comment_check", "submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.UnknownArgumentError):
                    getattr(client, method)(bad_argument=1, **self.common_kwargs)

    def test_protocol_error_comment_check(self):
        """
        ProtocolError is raised when ``comment_check()`` receives an unexpected
        response.

        """
        client = akismet.SyncClient(
            http_client=make_fixed_response_sync_client(response_text="bad"),
        )
        with self.assertRaises(akismet.ProtocolError):
            client.comment_check(**self.common_kwargs)

    def test_protocol_error_submit_ham_spam(self):
        """
        ProtocolError is raised when ``submit_ham()`` or ``submit_spam()`` receive an
        unexpected response.

        """
        client = akismet.SyncClient(
            http_client=make_fixed_response_sync_client(response_text="bad"),
        )
        for method in ("submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.ProtocolError):
                    getattr(client, method)(**self.common_kwargs)

    def test_protocol_error_verify_key(self):
        """
        ProtocolError is raised when ``verify_key()`` receives an unexpected response.

        """

        client = akismet.SyncClient(
            http_client=make_fixed_response_sync_client(response_text="bad"),
        )
        with self.assertRaises(akismet.ProtocolError):
            client.verify_key()



"""
Library version number.

"""

# SPDX-License-Identifier: BSD-3-Clause

# This is set here instead of in common.py or __init__.py because it's read dynamically
# by setuptools during the package build, and both common.py and __init__.py (which
# imports common.py) need to import the third-party httpx library, which is not
# available during package build (it only becomes available after package install).
LIBRARY_VERSION = "24.5.1"



"""
Legacy/deprecated Akismet client.

"""

# SPDX-License-Identifier: BSD-3-Clause

import os
import textwrap
import warnings
from typing import Optional

import httpx

from . import _common, _exceptions


class Akismet:
    """
    Legacy class-based Akismet client.

    This class is deprecated and will be removed in a future release. It receives no new
    features and only minimal security-oriented bugfixes. Please migrate as quickly as
    possible to one of the two supported client classes: akismet.SyncClient or
    akismet.AsyncClient.

    Two configuration parameters -- your Akismet API key and registered URL -- are
    required; they can be passed when instantiating, or set in the environment variables
    ``PYTHON_AKISMET_API_KEY`` and ``PYTHON_AKISMET_BLOG_URL``.

    The following operations of the Akismet API are exposed here:

    * :meth:`verify_key`

    * :meth:`comment_check`

    * :meth:`submit_spam`

    * :meth:`submit_ham`

    For full details of the Akismet API, see the Akismet documentation:

    https://akismet.com/development/api/#detailed-docs

    The verify_key operation will be automatically called for you as this class is
    instantiated; :exc:`~akismet.ConfigurationError` will be raised if the configuration
    cannot be found or if the supplied key/URL are invalid.

    """

    COMMENT_CHECK_URL = "https://rest.akismet.com/1.1/comment-check"
    SUBMIT_HAM_URL = "https://rest.akismet.com/1.1/submit-ham"
    SUBMIT_SPAM_URL = "https://rest.akismet.com/1.1/submit-spam"
    VERIFY_KEY_URL = "https://rest.akismet.com/1.1/verify-key"

    SUBMIT_SUCCESS_RESPONSE = "Thanks for making the web a better place."

    OPTIONAL_KEYS = [
        "blog_charset",
        "blog_lang",
        "comment_author",
        "comment_author_email",
        "comment_author_url",
        "comment_content",
        "comment_date_gmt",
        "comment_post_modified_gmt",
        "comment_type",
        "is_test",
        "permalink",
        "recheck_reason",
        "referrer",
        "user_role",
    ]

    user_agent_header = {"User-Agent": _common.USER_AGENT}

    def __init__(
        self,
        key: Optional[str] = None,
        blog_url: Optional[str] = None,
        http_client: Optional[httpx.Client] = None,
    ):
        warnings.warn(
            textwrap.dedent(
                """
            The akismet.Akismet API client is deprecated and will be removed in
            2025. Please migrate to either akismet.SyncClient or akismet.AsyncClient.
            """,
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        maybe_key = key if key is not None else os.getenv(_common._KEY_ENV_VAR, "")
        maybe_url = (
            blog_url if blog_url is not None else os.getenv(_common._URL_ENV_VAR, "")
        )
        if maybe_key == "" or maybe_url == "":
            raise _exceptions.ConfigurationError(
                textwrap.dedent(
                    f"""
                Could not find full Akismet configuration.

                Found API key: {maybe_key}
                Found blog URL: {maybe_url}
                """
                )
            )
        self.http_client = http_client or _common._get_sync_http_client()
        if not self.verify_key(maybe_key, maybe_url, http_client=self.http_client):
            raise _exceptions.APIKeyError(
                f"Akismet key ({maybe_key}, {maybe_url}) is invalid."
            )
        self.api_key = maybe_key
        self.blog_url = maybe_url

    def _api_request(
        self, endpoint: str, user_ip: str, user_agent: str, **kwargs: str
    ) -> httpx.Response:
        """
        Make a request to the Akismet API.

        This method is used for all API calls except key verification, since all
        endpoints other than key verification must interpolate the API key into the URL
        and supply certain basic data.

        """
        unknown_args = [k for k in kwargs if k not in self.OPTIONAL_KEYS]
        if unknown_args:
            raise _exceptions.UnknownArgumentError(
                "Unknown arguments while making request: {', '.join(unknown_args)}."
            )

        data = {
            "api_key": self.api_key,
            "blog": self.blog_url,
            "user_ip": user_ip,
            "user_agent": user_agent,
            **kwargs,
        }
        return self.http_client.post(endpoint, data=data)

    def _submission_request(  # pylint: disable=inconsistent-return-statements
        self, operation: str, user_ip: str, user_agent: str, **kwargs: str
    ) -> bool:
        """
        Submit spam or ham to the Akismet API.

        """
        endpoint = {
            "submit_spam": self.SUBMIT_SPAM_URL,
            "submit_ham": self.SUBMIT_HAM_URL,
        }[operation]
        response = self._api_request(endpoint, user_ip, user_agent, **kwargs)
        if response.text == self.SUBMIT_SUCCESS_RESPONSE:
            return True
        self._protocol_error(operation, response)

    @classmethod
    def _protocol_error(cls, operation: str, response: httpx.Response) -> None:
        """
        Raise an appropriate exception for unexpected API responses.

        """
        raise _exceptions.ProtocolError(
            textwrap.dedent(
                f"""
            Received unexpected or non-standard response from Akismet API.

            API operation was: {operation}
            API response received was: {response.text}
            Debug header value was: {response.headers.get('X-akismet-debug-help')}
            """
            )
        )

    @classmethod
    def verify_key(  # pylint: disable=inconsistent-return-statements
        cls, key: str, blog_url: str, http_client: Optional[httpx.Client] = None
    ) -> bool:
        """
        Verify an Akismet API key and URL.

        Returns :data:`True` if the key and URL are valid, :data:`False` otherwise.

        """
        if not blog_url.startswith(("http://", "https://")):
            raise _exceptions.ConfigurationError(
                textwrap.dedent(
                    f"""
                Invalid site URL specified: {blog_url}

                Akismet requires the full URL including the leading
                'http://' or 'https://'.
                """
                )
            )
        if http_client is None:  # pragma: no cover
            http_client = _common._get_sync_http_client()
        response = http_client.post(
            cls.VERIFY_KEY_URL,
            data={"key": key, "blog": blog_url},
        )
        if response.text == "valid":
            return True
        if response.text == "invalid":
            return False
        cls._protocol_error("verify_key", response)

    def comment_check(  # pylint: disable=inconsistent-return-statements
        self, user_ip: str, user_agent: str, **kwargs: str
    ) -> bool:
        """
        Check a comment to determine whether it is spam.

        The IP address and user-agent string of the remote user are required. All other
        arguments documented by Akismet (other than the PHP server information) are also
        optionally accepted.  See the Akismet API documentation for a full list:

        https://akismet.com/developers/comment-check/

        Like the Akismet web API, returns True for a comment that is spam, and False for
        a comment that is not spam.

        """
        response = self._api_request(
            self.COMMENT_CHECK_URL, user_ip, user_agent, **kwargs
        )
        if response.text == "true":
            return True
        if response.text == "false":
            return False
        self._protocol_error("comment_check", response)

    def submit_spam(self, user_ip: str, user_agent: str, **kwargs: str) -> bool:
        """
        Inform Akismet that a comment is spam.

        The IP address and user-agent string of the remote user are required. All other
        arguments documented by Akismet (other than the PHP server information) are also
        optionally accepted.  See the Akismet API documentation for a full list:

        https://akismet.com/developers/submit-spam-missed-spam/

        Returns True on success (the only expected response).

        """
        return self._submission_request("submit_spam", user_ip, user_agent, **kwargs)

    def submit_ham(self, user_ip: str, user_agent: str, **kwargs: str) -> bool:
        """
        Inform Akismet that a comment is not spam.

        The IP address and user-agent string of the remote user are required. All other
        arguments documented by Akismet (other than the PHP server information) are also
        optionally accepted.  See the Akismet API documentation for a full list:

        https://akismet.com/developers/submit-ham-false-positives/

        Returns True on success (the only expected response).

        """
        return self._submission_request("submit_ham", user_ip, user_agent, **kwargs)



"""
Test versions of the Akismet clients, for use both in Akismet's own testing and for
testing by applications and libraries which use Akismet.

"""

# SPDX-License-Identifier: BSD-3-Clause

from http import HTTPStatus
from typing import TYPE_CHECKING, Optional

import httpx

from . import _common
from ._async_client import AsyncClient
from ._sync_client import SyncClient

if TYPE_CHECKING:  # pragma: no cover
    import akismet

_TEST_KEY = "invalid-test-key"
_TEST_URL = "http://example.com/"

_COMMENT_CHECK_URL = f"/{_common._API_V11}/{_common._COMMENT_CHECK}"
_KEY_SITES_URL = f"/{_common._API_V12}/{_common._KEY_SITES}"
_SUBMIT_HAM_URL = f"/{_common._API_V11}/{_common._SUBMIT_HAM}"
_SUBMIT_SPAM_URL = f"/{_common._API_V11}/{_common._SUBMIT_SPAM}"
_USAGE_LIMIT_URL = f"/{_common._API_V12}/{_common._USAGE_LIMIT}"
_VERIFY_KEY_URL = f"/{_common._API_V11}/{_common._VERIFY_KEY}"

_COMMENT_CHECK_DISCARD_RESPONSE = {
    "content": "true",
    "headers": {"X-akismet-pro-tip": "discard"},
}
_COMMENT_CHECK_HAM_RESPONSE = {"content": "false"}
_COMMENT_CHECK_SPAM_RESPONSE = {"content": "true"}
_VERIFY_KEY_VALID_RESPONSE = {"content": "valid"}
_VERIFY_KEY_INVALID_RESPONSE = {"content": "invalid"}

_KEY_SITES_CSV = {
    # Sample CSV data taken from Akismet's dev docs.
    "content": """Active sites for 123YourAPIKey during 2022-09 (limit:10, offset: 0, total: 4)
Site,Total API Calls,Spam,Ham,Missed Spam,False Positives,Is Revoked
site6735.example.com,14446,33,13,0,9,false
site3026.example.com,8677,101,6,0,0,false
site3737.example.com,4230,65,5,2,0,true
site5653.example.com,2921,30,1,2,6,false"""
}
_KEY_SITES_JSON = {
    # Sample JSON from Akismet's dev docs.
    "json": {
        "2022-09": [
            {
                "site": "site6735.example.com",
                "api_calls": "2072",
                "spam": "2069",
                "ham": "3",
                "missed_spam": "0",
                "false_positives": "4",
                "is_revoked": False,
            },
            {
                "site": "site4748.example.com",
                "api_calls": "1633",
                "spam": "3",
                "ham": "1630",
                "missed_spam": "0",
                "false_positives": "0",
                "is_revoked": True,
            },
        ],
        "limit": 10,
        "offset": 0,
        "total": 2,
    }
}

_BASE_RESPONSE_MAP = {
    _COMMENT_CHECK_URL: _COMMENT_CHECK_SPAM_RESPONSE,
    _SUBMIT_HAM_URL: {"content": _common._SUBMISSION_RESPONSE},
    _SUBMIT_SPAM_URL: {"content": _common._SUBMISSION_RESPONSE},
    _USAGE_LIMIT_URL: {
        # Sample JSON from Akismet's dev docs.
        "json": {
            "limit": 350000,
            "usage": 7463,
            "percentage": "2.13",
            "throttled": False,
        }
    },
    _VERIFY_KEY_URL: _VERIFY_KEY_VALID_RESPONSE,
}


def _make_test_transport(
    comment_check_response: _common.CheckResponse = _common.CheckResponse.SPAM,
    verify_key_response: bool = True,
) -> httpx.MockTransport:
    """
    Build and return an ``httpx`` mock transport based on the given arguments:

    :param comment_check_response: How to classify submitted content.

    :param verify_key_response: Whether to treat the Akismet key/URL as valid.

    """
    custom_responses = {
        _COMMENT_CHECK_URL: {
            _common.CheckResponse.DISCARD: _COMMENT_CHECK_DISCARD_RESPONSE,
            _common.CheckResponse.HAM: _COMMENT_CHECK_HAM_RESPONSE,
            _common.CheckResponse.SPAM: _COMMENT_CHECK_SPAM_RESPONSE,
        }[comment_check_response],
        _VERIFY_KEY_URL: (
            _VERIFY_KEY_VALID_RESPONSE
            if verify_key_response
            else _VERIFY_KEY_INVALID_RESPONSE
        ),
    }
    response_map = {**_BASE_RESPONSE_MAP, **custom_responses}

    def _handler(request: httpx.Request) -> httpx.Response:
        """
        Mock transport handler which returns controlled responses.

        """
        if request.url.path != _KEY_SITES_URL:
            response_args = response_map[request.url.path]
        else:
            # key-sites is the only operation where anything other than the path
            # matters; it has a single query param which controls the response format,
            # so we have to adjust the response to match what was requested.
            response_args = (
                _KEY_SITES_CSV
                if request.url.query == b"format=csv"
                else _KEY_SITES_JSON
            )
        return httpx.Response(status_code=HTTPStatus.OK, **response_args)

    return httpx.MockTransport(_handler)


def _make_test_async_http_client(
    comment_check_response: _common.CheckResponse = _common.CheckResponse.SPAM,
    verify_key_response: bool = True,
) -> httpx.AsyncClient:
    """
    Construct and return an ``httpx.AsyncClient`` for testing purposes, based on the
    given arguments.

    :param comment_check_response: The desired comment-check response (ham, spam, or
       discard/"blatant spam").

    :param verify_key_response: The desired verify-key response: :data:`True` for a
       valid configuration, :data:`False` for invalid.

    """
    return httpx.AsyncClient(
        transport=_make_test_transport(comment_check_response, verify_key_response)
    )


def _make_test_sync_http_client(
    comment_check_response: _common.CheckResponse = _common.CheckResponse.SPAM,
    verify_key_response: bool = True,
) -> httpx.Client:
    """
    Construct and return an ``httpx.Client`` for testing purposes, based on the
    given arguments.

    :param comment_check_response: The desired comment-check response (ham, spam, or
       discard/"blatant spam").

    :param verify_key_response: The desired verify-key response: :data:`True` for a
       valid configuration, :data:`False` for invalid.

    """
    return httpx.Client(
        transport=_make_test_transport(comment_check_response, verify_key_response)
    )


class TestAsyncClient(AsyncClient):
    """
    A version of :class:`akismet.AsyncClient` for use in testing.

    This client exposes exactly the same API as :class:`~akismet.AsyncClient`, but will
    *not* make real requests to the Akismet web service. Instead it will return fixed
    responses for most Akismet API operations. The two configurable responses are
    comment-check and verify-key, which can be controlled by subclassing this class and
    setting the following attributes:

    .. attribute:: comment_check_response

       A value from the :class:`akismet.CheckResponse` enum, indicating the desired
       return value of the comment-check operation. Defaults to
       :attr:`~akismet.CheckResponse.SPAM` if not specified. Use this to test code which
       needs to handle the various possible responses from the Akismet API.

    .. attribute:: verify_key_response

       A :data:`bool` indicating the desired return value of the verify-key
       operation. Defaults to :data:`True` if not specified. Use this to test code which
       may need to handle or recover from invalid or missing Akismet configuration.

    Note that although you *can* pass the ``http_client`` argument to the constructors
    as with the base :class:`~akismet.AsyncClient`, this class will ignore that argument
    and instead always uses its own custom HTTP client to produce the desired fixed
    responses.

    For operations other than comment-check and verify-key, the response values will be:

    * submit-ham and submit-spam: :data:`True`

    * key-sites: the sample responses (depending on requested format) given in `the
      Akismet web service documentation for key-sites
      <https://akismet.com/developers/detailed-docs/key-sites-activity/>`_.

    * usage-limit: the sample response given in `the Akismet web service documentation
      for usage-limit <https://akismet.com/developers/detailed-docs/usage-limit/>`_.

    """

    comment_check_response: "akismet.CheckResponse" = _common.CheckResponse.HAM
    verify_key_response: bool = True

    def __init__(
        self,
        config: Optional["akismet.Config"] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ) -> None:
        """
        Default constructor.

        This overrides the behavior of the parent class to set up a custom HTTP client
        which returns fixed test responses. As a result, any explicit HTTP client passed
        in will be ignored.

        """
        test_client = _make_test_async_http_client(
            comment_check_response=self.comment_check_response,
            verify_key_response=self.verify_key_response,
        )
        super().__init__(config=config, http_client=test_client)


class TestSyncClient(SyncClient):
    """
    A version of :class:`akismet.SyncClient` for use in testing.

    This client exposes exactly the same API as :class:`~akismet.SyncClient`, but will
    *not* make real requests to the Akismet web service. Instead it will return fixed
    responses for most Akismet API operations. The two configurable responses are
    comment-check and verify-key, which can be controlled by subclassing this class and
    setting the following attributes:

    .. attribute:: comment_check_response

       A value from the :class:`akismet.CheckResponse` enum, indicating the desired
       return value of the comment-check operation. Defaults to
       :attr:`~akismet.CheckResponse.SPAM` if not specified. Use this to test code which
       needs to handle the various possible responses from the Akismet API.

    .. attribute:: verify_key_response

       A :data:`bool` indicating the desired return value of the verify-key
       operation. Defaults to :data:`True` if not specified. Use this to test code which
       may need to handle or recover from invalid or missing Akismet configuration.

    Note that although you *can* pass the ``http_client`` argument to the constructors
    as with the base :class:`~akismet.SyncClient`, this class will ignore that argument
    and instead always uses its own custom HTTP client to produce the desired fixed
    responses.

    For operations other than comment-check and verify-key, the response values will be:

    * submit-ham and submit-spam: :data:`True`

    * key-sites: the sample responses (depending on requested format) given in `the
      Akismet web service documentation for key-sites
      <https://akismet.com/developers/detailed-docs/key-sites-activity/>`_.

    * usage-limit: the sample response given in `the Akismet web service documentation
      for usage-limit <https://akismet.com/developers/detailed-docs/usage-limit/>`_.

    """

    comment_check_response: "akismet.CheckResponse" = _common.CheckResponse.HAM
    verify_key_response: bool = True

    def __init__(
        self,
        config: Optional["akismet.Config"] = None,
        http_client: Optional[httpx.Client] = None,
    ) -> None:
        """
        Default constructor.

        This overrides the behavior of the parent class to set up a custom HTTP client
        which returns fixed test responses. As a result, any explicit HTTP client passed
        in will be ignored.

        """
        test_client = _make_test_sync_http_client(
            comment_check_response=self.comment_check_response,
            verify_key_response=self.verify_key_response,
        )
        super().__init__(config=config, http_client=test_client)



"""
A Python interface to `the Akismet spam-filtering service <https://akismet.com>`_.

Two Akismet API clients are available from this library:

* :class:`akismet.SyncClient` is an Akismet API client which performs synchronous
  (blocking) HTTP requests to the Akismet web service.

* :class:`akismet.AsyncClient` is an Akismet API client which performs asynchronous
  (``async``/``await``/non-blocking) HTTP requests to the Akismet web service.

Aside from one being sync and the other async, the two clients expose identical APIs,
and implement all methods of `the Akismet web API <https://akismet.com/developers/>`_.

To use this library, you will need to obtain an Akismet API key and register a site for
use with the Akismet web service; you can do this at <https://akismet.com>. Once you
have a key and corresponding registered site URL to use with it, place them in the
environment variables ``PYTHON_AKISMET_API_KEY`` and ``PYTHON_AKISMET_BLOG_URL``, and
they will be automatically detected and used.

You can then construct a client instance and call its methods. For creating a long-lived
API client instance, it's recommended that you use the ``validated_client()``
constructor method, which will automatically validate your API key with the Akismet web
service. For example, to check a submitted forum post for spam:

.. code-block:: python

   import akismet

   akismet_client = akismet.SyncClient.validated_client()

   if akismet_client.comment_check(
       user_ip=submitter_ip,
       comment_content=submitted_content,
       comment_type="forum-post",
       comment_author=submitter_name
   ):
       # This piece of content was classified as spam; handle it appropriately.

Or using the asynchronous client:

.. code-block:: python

   import akismet

   akismet_client = await akismet.AsyncClient.validated_client()

   if await akismet_client.comment_check(
       user_ip=submitter_ip,
       comment_content=submitted_content,
       comment_type="forum-post",
       comment_author=submitter_name
   ):
       # This piece of content was classified as spam; handle it appropriately.

You can also use either client class as a context manager. This does *not* require the
``validated_client()`` constructor, because your API key is validated on entering the
``with`` block.

.. code-block:: python

   import akismet

   with akismet.SyncClient() as akismet_client:
       if akismet_client.comment_check(
           user_ip=submitter_ip,
           comment_content=submitted_content,
           comment_type="forum-post",
           comment_author=submitter_name
       ):
           # This piece of content was classified as spam; handle it appropriately.

Or using the asynchronous client:

.. code-block:: python

   import akismet

   async with akismet.AsyncClient() as akismet_client:
       if await akismet_client.comment_check(
           user_ip=submitter_ip,
           comment_content=submitted_content,
           comment_type="forum-post",
           comment_author=submitter_name
       ):
           # This piece of content was classified as spam; handle it appropriately.

"""

# SPDX-License-Identifier: BSD-3-Clause

from ._async_client import AsyncClient
from ._common import USER_AGENT, CheckResponse, Config
from ._exceptions import (
    AkismetError,
    APIKeyError,
    ConfigurationError,
    ProtocolError,
    RequestError,
    UnknownArgumentError,
)
from ._legacy_client import Akismet
from ._sync_client import SyncClient
from ._test_clients import TestAsyncClient, TestSyncClient
from ._version import LIBRARY_VERSION

__version__ = LIBRARY_VERSION

__all__ = [
    "APIKeyError",
    "Akismet",
    "AkismetError",
    "AsyncClient",
    "CheckResponse",
    "Config",
    "ConfigurationError",
    "ProtocolError",
    "RequestError",
    "SyncClient",
    "TestAsyncClient",
    "TestSyncClient",
    "UnknownArgumentError",
    "USER_AGENT",
]



"""
Base test class and utilities.

"""

# SPDX-License-Identifier: BSD-3-Clause

import os
import typing
import unittest
from http import HTTPStatus

import httpx

import akismet
from akismet import _common, _test_clients


def make_fixed_response_transport(
    response_text: str = "true",
    status_code: HTTPStatus = HTTPStatus.OK,
    response_json: typing.Optional[dict] = None,
) -> httpx.MockTransport:
    """
    Return an ``httpx`` transport that produces a fixed response, for use
    in testing.

    The transport will return a response consisting of:

    * ``status_code`` (default 200)
    * ``response_json`` as the JSON content, if supplied
    * Otherwise ``response_text`` (default ``"true"``) as the response text

    """

    def _handler(
        request: httpx.Request,  # pylint: disable=unused-argument
    ) -> httpx.Response:
        """
        Mock transport handler which returns a controlled response.

        """
        response_kwargs = {"status_code": status_code, "content": response_text}
        if response_json is not None:
            del response_kwargs["content"]
            response_kwargs["json"] = response_json
        return httpx.Response(**response_kwargs)

    return httpx.MockTransport(_handler)


class CommonData:  # pylint: disable=too-few-public-methods
    """
    Common data for all Akismet tests.

    """

    api_key = os.getenv("PYTHON_AKISMET_API_KEY")
    site_url = os.getenv("PYTHON_AKISMET_BLOG_URL")
    verify_key_url = f"{_common._API_URL}/{_common._API_V11}/{_common._VERIFY_KEY}"

    config = akismet.Config(key=_test_clients._TEST_KEY, url=_test_clients._TEST_URL)
    common_kwargs = {"user_ip": "127.0.0.1"}


class AkismetTests(CommonData, unittest.TestCase):
    """
    Base class for synchronous Akismet tests.

    """


class AsyncAkismetTests(CommonData, unittest.IsolatedAsyncioTestCase):
    """
    Base class for asynchronous Akismet tests.

    """



"""
Tests for the Akismet test clients.

"""

from akismet import APIKeyError, CheckResponse, TestAsyncClient, TestSyncClient

from .base import AkismetTests, AsyncAkismetTests


class SyncTestClientTests(AkismetTests):
    """
    Tests for the synchronous Akismet test client.

    """

    def test_comment_check_default_not_spam(self):
        """
        The default configuration of the test client marks all content as non-spam.

        """
        client = TestSyncClient(config=self.config)
        assert (
            client.comment_check(comment_content="Test", **self.common_kwargs)
            == CheckResponse.HAM
        )

    def test_comment_check_configuration(self):
        """
        Setting comment_check_response explicitly will cause the test client to mark
        content accordingly.

        """
        for response_value in CheckResponse:
            with self.subTest(comment_check_response=response_value):

                class _Client(TestSyncClient):
                    """
                    Test client with explicit comment_check_response.

                    """

                    comment_check_response = response_value

                client = _Client(config=self.config)
                assert (
                    client.comment_check(comment_content="Test", **self.common_kwargs)
                    == response_value
                )

    def test_verify_key_default(self):
        """
        The default configuration of the test client succeeds at key verification.

        """
        client = TestSyncClient(config=self.config)
        assert client.verify_key()

    def test_verify_key_explicit_success(self):
        """
        Setting verify_key_response explicitly to True will cause the test client to
        succeed at key verification.

        """

        class _Client(TestSyncClient):
            """
            Test client with explicit verify_key_response.

            """

            verify_key_response = True

        # Explicit configuration succeeds.
        client = _Client(config=self.config)
        assert client.verify_key()

        # Implicit configuration succeeds.
        _Client.validated_client()

    def test_verify_key_explicit_failure(self):
        """
        Setting verify_key_response explicitly to False will cause the test client
        to fail at key verification.

        """

        class _Client(TestSyncClient):
            """
            Test client with explicit verify_key_response.

            """

            verify_key_response = False

        # Explicit configuration fails.
        client = _Client(config=self.config)
        assert not client.verify_key()

        # Implicit configuration fails.
        with self.assertRaises(APIKeyError):
            _Client.validated_client()


class AsyncTestClientTests(AsyncAkismetTests):
    """
    Tests for the asynchronous Akismet test client.

    """

    async def test_comment_check_default_not_spam(self):
        """
        The default configuration of the test client marks all content as non-spam.

        """
        client = TestAsyncClient(config=self.config)
        assert (
            await client.comment_check(comment_content="Test", **self.common_kwargs)
            == CheckResponse.HAM
        )

    async def test_comment_check_configuration(self):
        """
        Setting comment_check_response explicitly will cause the test client to mark
        content accordingly.

        """
        for response_value in CheckResponse:
            with self.subTest(comment_check_response=response_value):

                class _Client(TestAsyncClient):
                    """
                    Test client with explicit comment_check_response.

                    """

                    comment_check_response = response_value

                client = _Client(config=self.config)
                assert (
                    await client.comment_check(
                        comment_content="Test", **self.common_kwargs
                    )
                    == response_value
                )

    async def test_verify_key_default(self):
        """
        The default configuration of the test client succeeds at key verification.

        """
        client = TestAsyncClient(config=self.config)
        assert await client.verify_key()

    async def test_verify_key_explicit_success(self):
        """
        Setting verify_key_response explicitly to True will cause the test client to
        succeed at key verification.

        """

        class _Client(TestAsyncClient):
            """
            Test client with explicit verify_key_response.

            """

            verify_key_response = True

        # Explicit configuration succeeds.
        client = _Client(config=self.config)
        assert await client.verify_key()

        # Implicit configuration succeeds.
        await _Client.validated_client()

    async def test_verify_key_explicit_failure(self):
        """
        Setting verify_key_response explicitly to False will cause the test client
        to fail at key verification.

        """

        class _Client(TestAsyncClient):
            """
            Test client with explicit verify_key_response.

            """

            verify_key_response = False

        # Explicit configuration fails.
        client = _Client(config=self.config)
        assert not await client.verify_key()

        # Implicit configuration fails.
        with self.assertRaises(APIKeyError):
            await _Client.validated_client()



"""
End-to-end tests of the Akismet API clients, which will make real requests to the
Akismet web service.

"""

# SPDX-License-Identifier: BSD-3-Clause

import os

import akismet
from akismet import _common

from .base import AkismetTests, AsyncAkismetTests

BAD_KEY = "INVALID_TEST_KEY"
BAD_URL = "http://example.com"

# The Akismet web API guarantees that these values will trigger certain types of
# responses.
HAM_ROLE = "administrator"
SPAM_AUTHOR = "akismet-guaranteed-spam"


class SyncAkismetEndToEndTests(AkismetTests):
    """
    End-to-end tests of the synchronous Akismet API client.

    """

    common_kwargs = {"user_ip": "127.0.0.1", "is_test": 1}

    def setUp(self):
        """
        Create a default client instance for use in testing.

        """
        self.client = akismet.SyncClient.validated_client()

    def test_construct_config_valid(self):
        """
        With a valid configuration, constructing a client succeeds.

        """
        akismet.SyncClient.validated_client()

    def test_construct_config_invalid_key(self):
        """
        With an invalid API key, constructing a client raises an APIKeyError.

        """

        try:
            os.environ[_common._KEY_ENV_VAR] = BAD_KEY
            os.environ[_common._URL_ENV_VAR] = BAD_URL
            with self.assertRaises(akismet.APIKeyError):
                akismet.SyncClient.validated_client()
        finally:
            os.environ[_common._KEY_ENV_VAR] = self.api_key
            os.environ[_common._URL_ENV_VAR] = self.site_url

    def test_verify_key_valid(self):
        """
        ``verify_key()`` returns True when the config is valid.

        """
        assert self.client.verify_key()

    def test_verify_key_invalid(self):
        """
        ``verify_key()`` returns False when the config is invalid.

        """
        assert not self.client.verify_key(BAD_KEY, BAD_URL)

    def test_request_with_invalid_key(self):
        """
        The request methods other than ``verify_key()`` raise
        ``akismet.APIKeyError`` if called with an invalid API key/URL.

        """
        client = akismet.SyncClient(
            config=self.config,
        )
        for method in ("comment_check", "submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.APIKeyError):
                    getattr(client, method)(**self.common_kwargs)
        for method in ("key_sites", "usage_limit"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.APIKeyError):
                    getattr(client, method)()

    def test_comment_check_spam(self):
        """
        ``comment_check()`` returns the SPAM value when Akismet declares the content
        to be spam.

        """
        assert (
            self.client.comment_check(
                comment_content="test", comment_author=SPAM_AUTHOR, **self.common_kwargs
            )
            == akismet.CheckResponse.SPAM
        )

    def test_comment_check_ham(self):
        """
        ``comment_check()`` returns the HAM value when Akismet declares the content
        to be ham.

        """
        assert (
            self.client.comment_check(
                comment_content="test", user_role=HAM_ROLE, **self.common_kwargs
            )
            == akismet.CheckResponse.HAM
        )

    def test_submit_ham(self):
        """
        ``submit_ham()`` returns True when Akismet accepts the submission.

        """
        assert self.client.submit_ham(
            comment_content="test", user_role=HAM_ROLE, **self.common_kwargs
        )

    def test_submit_spam(self):
        """
        ``submit_spam()`` returns True when Akismet accepts the submission.

        """
        assert self.client.submit_spam(
            comment_content="test", comment_author=SPAM_AUTHOR, **self.common_kwargs
        )


class AsyncAkismetEndToEndTests(AsyncAkismetTests):
    """
    End-to-end tests of the asynchronous Akismet API client.

    """

    common_kwargs = {"user_ip": "127.0.0.1", "is_test": 1}

    async def asyncSetUp(self):
        """
        Create a default client instance for use in testing.

        """
        self.client = await akismet.AsyncClient.validated_client()

    async def test_construct_config_valid(self):
        """
        With a valid configuration, constructing a client succeeds.

        """
        await akismet.AsyncClient.validated_client()

    async def test_construct_config_invalid_key(self):
        """
        With an invalid API key, constructing a client raises an APIKeyError.

        """

        try:
            os.environ[_common._KEY_ENV_VAR] = BAD_KEY
            os.environ[_common._URL_ENV_VAR] = BAD_URL
            with self.assertRaises(akismet.APIKeyError):
                await akismet.AsyncClient.validated_client()
        finally:
            os.environ[_common._KEY_ENV_VAR] = self.api_key
            os.environ[_common._URL_ENV_VAR] = self.site_url

    async def test_verify_key_valid(self):
        """
        ``verify_key()`` returns True when the config is valid.

        """
        assert await self.client.verify_key()

    async def test_verify_key_invalid(self):
        """
        ``verify_key()`` returns False when the config is invalid.

        """
        assert not await self.client.verify_key(BAD_KEY, BAD_URL)

    async def test_request_with_invalid_key(self):
        """
        The HTTP POST request methods -- ``comment_check()``, ``submit_ham()``, and
        ``submit_spam()`` -- raise ``APIKeyError`` when called with an invalid API key
        and/or site URL.

        """
        client = akismet.AsyncClient(
            config=self.config,
        )
        for method in ("comment_check", "submit_ham", "submit_spam"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.APIKeyError):
                    await getattr(client, method)(**self.common_kwargs)
        for method in ("key_sites", "usage_limit"):
            with self.subTest(method=method):
                with self.assertRaises(akismet.APIKeyError):
                    await getattr(client, method)()

    async def test_comment_check_spam(self):
        """
        ``comment_check()`` returns the SPAM value when Akismet declares the content
        to be spam.

        """
        assert (
            await self.client.comment_check(
                comment_content="test", comment_author=SPAM_AUTHOR, **self.common_kwargs
            )
            == akismet.CheckResponse.SPAM
        )

    async def test_comment_check_ham(self):
        """
        ``comment_check()`` returns the HAM value when Akismet declares the content
        to be ham.

        """
        assert (
            await self.client.comment_check(
                comment_content="test", user_role=HAM_ROLE, **self.common_kwargs
            )
            == akismet.CheckResponse.HAM
        )

    async def test_submit_ham(self):
        """
        ``submit_ham()`` returns True when Akismet accepts the submission.

        """
        assert await self.client.submit_ham(
            comment_content="test", user_role=HAM_ROLE, **self.common_kwargs
        )

    async def test_submit_spam(self):
        """
        ``submit_spam()`` returns True when Akismet accepts the submission.

        """
        assert await self.client.submit_spam(
            comment_content="test", comment_author=SPAM_AUTHOR, **self.common_kwargs
        )


class LegacyAkismetEndToEndTests(AkismetTests):
    """
    End-to-end tests of the legacy/deprecated Akismet API client.

    """

    common_kwargs = {"user_ip": "127.0.0.1", "user_agent": "Mozilla", "is_test": 1}

    def setUp(self):
        """
        Create a default client instance for use in testing.

        """
        self.client = akismet.Akismet(key=self.api_key, blog_url=self.site_url)

    def test_construct_config_valid(self):
        """
        With a valid configuration, constructing a client succeeds.

        """
        akismet.Akismet(key=self.api_key, blog_url=self.site_url)

    def test_construct_config_invalid_key(self):
        """
        With an invalid API key, constructing a client raises an APIKeyError.

        """
        with self.assertRaises(akismet.APIKeyError):
            akismet.Akismet(key=BAD_KEY, blog_url=BAD_URL)

    def test_verify_key_valid(self):
        """
        ``verify_key()`` returns True when the config is valid.

        """
        assert self.client.verify_key(self.api_key, self.site_url)

    def test_verify_key_invalid(self):
        """
        ``verify_key()`` returns False when the config is invalid.

        """
        assert not self.client.verify_key(BAD_KEY, BAD_URL)

    def test_comment_check_spam(self):
        """
        ``comment_check()`` returns the SPAM value when Akismet declares the content
        to be spam.

        """
        assert self.client.comment_check(
            comment_content="test", comment_author=SPAM_AUTHOR, **self.common_kwargs
        )

    def test_comment_check_ham(self):
        """
        ``comment_check()`` returns the HAM value when Akismet declares the content
        to be ham.

        """
        assert not self.client.comment_check(
            comment_content="test", user_role=HAM_ROLE, **self.common_kwargs
        )

    def test_submit_ham(self):
        """
        ``submit_ham()`` returns True when Akismet accepts the submission.

        """
        assert self.client.submit_ham(
            comment_content="test", user_role=HAM_ROLE, **self.common_kwargs
        )

    def test_submit_spam(self):
        """
        ``submit_spam()`` returns True when Akismet accepts the submission.

        """
        assert self.client.submit_spam(
            comment_content="test", comment_author=SPAM_AUTHOR, **self.common_kwargs
        )



"""
Automated testing via nox (https://nox.thea.codes/).

Combined with a working installation of nox (see ``nox`` documentation), this file
specifies a matrix of tests, linters, and other quality checks which can be run
individually or as a suite.

To see available tasks, run ``python -m nox --list``. To run all available tasks --
which requires functioning installs of all supported Python versions -- run ``python -m
nox``. To run a single task, use ``python -m nox --session`` with the name of that task.

"""

import os
import pathlib
import shutil
import typing

import nox

nox.options.default_venv_backend = "venv"
nox.options.keywords = "not release"
nox.options.reuse_existing_virtualenvs = True

PACKAGE_NAME = "akismet"

NOXFILE_PATH = pathlib.Path(__file__).parents[0]
ARTIFACT_PATHS = (
    NOXFILE_PATH / "src" / f"{PACKAGE_NAME}.egg-info",
    NOXFILE_PATH / "build",
    NOXFILE_PATH / "dist",
    NOXFILE_PATH / "__pycache__",
    NOXFILE_PATH / "src" / "__pycache__",
    NOXFILE_PATH / "src" / PACKAGE_NAME / "__pycache__",
    NOXFILE_PATH / "tests" / "__pycache__",
)

TEST_KEY = "INVALID_TEST_KEY"
TEST_URL = "http://example.com"


def clean(paths: typing.Iterable[os.PathLike] = ARTIFACT_PATHS) -> None:
    """
    Clean up after a test run.

    """
    [
        shutil.rmtree(path) if path.is_dir() else path.unlink()
        for path in paths
        if path.exists()
    ]


# Tasks which run the package's test suites.
# -----------------------------------------------------------------------------------


@nox.session(python=["3.8", "3.9", "3.10", "3.11", "3.12"], tags=["tests"])
def tests_with_coverage(session: nox.Session) -> None:
    """
    Run the package's unit tests, with coverage report.

    """
    session.install(".[tests]")
    session.run(
        f"python{session.python}",
        "-Wonce::DeprecationWarning",
        "-Im",
        "coverage",
        "run",
        "--source",
        PACKAGE_NAME,
        "-m",
        "unittest",
        "discover",
        env={"PYTHON_AKISMET_API_KEY": TEST_KEY, "PYTHON_AKISMET_BLOG_URL": TEST_URL},
    )
    session.run(
        f"python{session.python}",
        "-Im",
        "coverage",
        "report",
        "--show-missing",
    )
    clean()


@nox.session(python=["3.8", "3.9", "3.10", "3.11", "3.12"], tags=["tests", "release"])
def tests_end_to_end(session: nox.Session) -> None:
    """
    Run the end-to-end (live Akismet API) tests.

    """
    session.install(".[tests]")
    session.run(
        f"python{session.python}",
        "-Wonce::DeprecationWarning",
        "-Im",
        "unittest",
        "discover",
        "--pattern",
        "end_to_end*",
        env={
            "PYTHON_AKISMET_API_KEY": os.getenv("PYTHON_AKISMET_API_KEY", ""),
            "PYTHON_AKISMET_BLOG_URL": os.getenv("PYTHON_AKISMET_BLOG_URL", ""),
        },
    )
    clean()


# Tasks which test the package's documentation.
# -----------------------------------------------------------------------------------


@nox.session(python=["3.12"], tags=["docs"])
def docs_build(session: nox.Session) -> None:
    """
    Build the package's documentation as HTML.

    """
    session.install(".[docs]")
    session.chdir("docs")
    session.run(
        f"python{session.python}",
        "-Im",
        "sphinx",
        "-b",
        "html",
        "-d",
        f"{session.bin}/../tmp/doctrees",
        ".",
        f"{session.bin}/../tmp/html",
    )
    clean()


@nox.session(python=["3.12"], tags=["docs"])
def docs_docstrings(session: nox.Session) -> None:
    """
    Enforce the presence of docstrings on all modules, classes, functions, and
    methods.

    """
    # interrogate implicitly depends on pkg_resources, which is part of setuptools but
    # as of Python 3.12, the venv module no longer automatically installed setuptools
    # into newly-created environments. So we install it manually here.
    session.install("interrogate", "setuptools")
    session.run(f"python{session.python}", "-Im", "interrogate", "--version")
    session.run(
        f"python{session.python}",
        "-Im",
        "interrogate",
        "-v",
        "src/",
        "tests/",
        "noxfile.py",
    )
    clean()


@nox.session(python=["3.12"], tags=["docs"])
def docs_spellcheck(session: nox.Session) -> None:
    """
    Spell-check the package's documentation.

    """
    session.install(
        "pyenchant",
        "sphinxcontrib-spelling",
        ".[docs]",
    )
    build_dir = session.create_tmp()
    session.chdir("docs")
    session.run(
        f"python{session.python}",
        "-Im",
        "sphinx",
        "-W",  # Promote warnings to errors, so that misspelled words fail the build.
        "-b",
        "spelling",
        "-d",
        f"{build_dir}/doctrees",
        ".",
        f"{build_dir}/html",
        # On Apple Silicon Macs, this environment variable needs to be set so
        # pyenchant can find the "enchant" C library. See
        # https://github.com/pyenchant/pyenchant/issues/265#issuecomment-1126415843
        env={"PYENCHANT_LIBRARY_PATH": os.getenv("PYENCHANT_LIBRARY_PATH", "")},
    )
    clean()


# Code formatting checks.
#
# These checks do *not* reformat code -- that happens in pre-commit hooks -- but will
# fail a CI build if they find any code that needs reformatting.
# -----------------------------------------------------------------------------------


@nox.session(python=["3.12"], tags=["formatters"])
def format_black(session: nox.Session) -> None:
    """
    Check code formatting with Black.

    """
    session.install("black>=24.0,<25.0")
    session.run(f"python{session.python}", "-Im", "black", "--version")
    session.run(
        f"python{session.python}",
        "-Im",
        "black",
        "--check",
        "--diff",
        "src/",
        "tests/",
        "docs/",
        "noxfile.py",
    )
    clean()


@nox.session(python=["3.12"], tags=["formatters"])
def format_isort(session: nox.Session) -> None:
    """
    Check import order with isort.

    """
    session.install("isort")
    session.run(f"python{session.python}", "-Im", "isort", "--version")
    session.run(
        f"python{session.python}",
        "-Im",
        "isort",
        "--check-only",
        "--diff",
        "src/",
        "tests/",
        "docs/",
        "noxfile.py",
    )
    clean()


# Linters.
# -----------------------------------------------------------------------------------


@nox.session(python=["3.12"], tags=["linters", "security"])
def lint_bandit(session: nox.Session) -> None:
    """
    Lint code with the Bandit security analyzer.

    """
    session.install("bandit[toml]")
    session.run(f"python{session.python}", "-Im", "bandit", "--version")
    session.run(
        f"python{session.python}",
        "-Im",
        "bandit",
        "-c",
        "./pyproject.toml",
        "-r",
        "src/",
        "tests/",
    )
    clean()


@nox.session(python=["3.12"], tags=["linters"])
def lint_flake8(session: nox.Session) -> None:
    """
    Lint code with flake8.

    """
    session.install("flake8", "flake8-bugbear")
    session.run(f"python{session.python}", "-Im", "flake8", "--version")
    session.run(
        f"python{session.python}",
        "-Im",
        "flake8",
        "src/",
        "tests/",
        "docs/",
        "noxfile.py",
    )
    clean()


@nox.session(python=["3.12"], tags=["linters"])
def lint_pylint(session: nox.Session) -> None:
    """
    Lint code with Pylint.

    """
    # Pylint requires that all dependencies be importable during the run. This package
    # does not have any direct dependencies, nor does the normal test suite, but the
    # full conformance suite does require a few extra libraries, so they're installed
    # here.
    session.install("httpx", "pylint")
    session.run(f"python{session.python}", "-Im", "pylint", "--version")
    session.run(f"python{session.python}", "-Im", "pylint", "src/", "tests/")
    clean()


# Packaging checks.
# -----------------------------------------------------------------------------------


@nox.session(python=["3.12"], tags=["packaging"])
def package_build(session: nox.Session) -> None:
    """
    Check that the package builds.

    """
    session.install("build")
    session.run(f"python{session.python}", "-Im", "build", "--version")
    session.run(f"python{session.python}", "-Im", "build")
    clean()


@nox.session(python=["3.12"], tags=["packaging"])
def package_description(session: nox.Session) -> None:
    """
    Check that the package description will render on the Python Package Index.

    """
    package_dir = session.create_tmp()
    session.install("build", "twine")
    session.run(f"python{session.python}", "-Im", "build", "--version")
    session.run(f"python{session.python}", "-Im", "twine", "--version")
    session.run(
        f"python{session.python}",
        "-Im",
        "build",
        "--wheel",
        "--outdir",
        f"{package_dir}/build",
    )
    session.run(
        f"python{session.python}", "-Im", "twine", "check", f"{package_dir}/build/*"
    )
    clean()


@nox.session(python=["3.12"], tags=["packaging"])
def package_manifest(session: nox.Session) -> None:
    """
    Check that the set of files in the package matches the set under version control.

    """
    session.install("check-manifest")
    session.run(f"python{session.python}", "-Im", "check_manifest", "--version")
    session.run(f"python{session.python}", "-Im", "check_manifest", "--verbose")
    clean()


@nox.session(python=["3.12"], tags=["packaging"])
def package_pyroma(session: nox.Session) -> None:
    """
    Check package quality with pyroma.

    """
    session.install("pyroma")
    session.run(f"python{session.python}", "-Im", "pyroma", ".")
    clean()


@nox.session(python=["3.12"], tags=["packaging"])
def package_wheel(session: nox.Session) -> None:
    """
    Check the built wheel package for common errors.

    """
    package_dir = session.create_tmp()
    session.install("build", "check-wheel-contents")
    session.run(f"python{session.python}", "-Im", "build", "--version")
    session.run(f"python{session.python}", "-Im", "check_wheel_contents", "--version")
    session.run(
        f"python{session.python}",
        "-Im",
        "build",
        "--wheel",
        "--outdir",
        f"{package_dir}/build",
    )
    session.run(
        f"python{session.python}", "-Im", "check_wheel_contents", f"{package_dir}/build"
    )
    clean()



"""
Tests for the deprectaed legacy Akismet API client.

"""

# SPDX-License-Identifier: BSD-3-Clause

import os

import akismet
from akismet import _common, _test_clients

from . import base
from .test_sync_client import make_fixed_response_sync_client


class LegacyAkismetConfigurationTests(base.AkismetTests):
    """
    Tests configuration of the legacy Akismet API class.

    """

    base_kwargs = {
        "user_ip": "127.0.0.1",
        "user_agent": "Mozilla",
    }

    def test_config_from_args(self):
        """
        Configuring via explicit arguments succeeds.

        """
        api = akismet.Akismet(
            key=self.api_key,
            blog_url=self.site_url,
            http_client=_test_clients._make_test_sync_http_client(),
        )
        self.assertEqual(self.api_key, api.api_key)
        self.assertEqual(self.site_url, api.blog_url)

    def test_bad_config_args(self):
        """
        Configuring with bad arguments fails.

        """
        with self.assertRaises(akismet.APIKeyError):
            akismet.Akismet(
                key="invalid",
                blog_url="http://invalid",
                http_client=_test_clients._make_test_sync_http_client(
                    verify_key_response=False
                ),
            )

    def test_config_from_env(self):
        """
        Configuring via environment variables succeeds.

        """
        api = akismet.Akismet(
            key=None,
            blog_url=None,
            http_client=_test_clients._make_test_sync_http_client(),
        )
        self.assertEqual(self.api_key, api.api_key)
        self.assertEqual(self.site_url, api.blog_url)

        api = akismet.Akismet(http_client=_test_clients._make_test_sync_http_client())
        self.assertEqual(self.api_key, api.api_key)
        self.assertEqual(self.site_url, api.blog_url)

    def test_bad_config_env(self):
        """
        Configuring with bad environment variables fails.

        """
        try:
            os.environ[_common._KEY_ENV_VAR] = "invalid"
            os.environ[_common._URL_ENV_VAR] = "http://invalid"
            with self.assertRaises(akismet.APIKeyError):
                akismet.Akismet(
                    http_client=_test_clients._make_test_sync_http_client(
                        verify_key_response=False
                    )
                )
        finally:
            os.environ[_common._KEY_ENV_VAR] = self.api_key
            os.environ[_common._URL_ENV_VAR] = self.site_url

    def test_bad_config_missing_key(self):
        """
        Configuring with missing key fails.

        """
        try:
            del os.environ[_common._KEY_ENV_VAR]
            with self.assertRaises(akismet.ConfigurationError):
                akismet.Akismet(
                    http_client=_test_clients._make_test_sync_http_client(
                        verify_key_response=False
                    )
                )
        finally:
            os.environ[_common._KEY_ENV_VAR] = self.api_key
            os.environ[_common._URL_ENV_VAR] = self.site_url

    def test_bad_config_missing_url(self):
        """
        Configuring with missing URL fails.

        """
        try:
            del os.environ[_common._URL_ENV_VAR]
            with self.assertRaises(akismet.ConfigurationError):
                akismet.Akismet(
                    http_client=_test_clients._make_test_sync_http_client(
                        verify_key_response=False
                    )
                )
        finally:
            os.environ[_common._URL_ENV_VAR] = self.site_url

    def test_bad_url(self):
        """
        Configuring with a bad URL fails.

        """
        bad_urls = (
            "example.com",
            "ftp://example.com",
            "www.example.com",
            "http//example.com",
            "https//example.com",
        )
        for url in bad_urls:
            with self.assertRaises(akismet.ConfigurationError):
                akismet.Akismet(
                    key=self.api_key,
                    blog_url=url,
                    http_client=_test_clients._make_test_sync_http_client(
                        verify_key_response=False
                    ),
                )

    def test_missing_config(self):
        """
        Instantiating without any configuration fails.

        """
        with self.assertRaises(akismet.ConfigurationError):
            akismet.Akismet(
                key=None,
                blog_url=None,
                http_client=_test_clients._make_test_sync_http_client(
                    verify_key_response=False
                ),
            )
        with self.assertRaises(akismet.ConfigurationError):
            akismet.Akismet(
                http_client=_test_clients._make_test_sync_http_client(
                    verify_key_response=False
                )
            )

    def test_user_agent(self):
        """
        The Akismet class creates the correct user-agent string.

        """
        api = akismet.Akismet(
            key=self.api_key,
            blog_url=self.site_url,
            http_client=_test_clients._make_test_sync_http_client(),
        )
        self.assertEqual(api.user_agent_header["User-Agent"], _common.USER_AGENT)


class LegacyAkismetAPITests(base.AkismetTests):
    """
    Tests implementation of the legacy Akismet API.

    """

    base_kwargs = {
        "user_ip": "127.0.0.1",
        "user_agent": "Mozilla",
    }

    def test_verify_key_valid(self):
        """
        The verify_key operation succeeds with a valid key and URL.

        """
        self.assertTrue(
            akismet.Akismet.verify_key(
                self.api_key,
                self.site_url,
                http_client=_test_clients._make_test_sync_http_client(),
            )
        )

    def test_verify_key_invalid(self):
        """
        The verify_key operation fails with an invalid key and URL.

        """
        self.assertFalse(
            akismet.Akismet.verify_key(
                "invalid",
                "http://invalid",
                http_client=_test_clients._make_test_sync_http_client(
                    verify_key_response=False
                ),
            )
        )

    def test_comment_check_spam(self):
        """
        The comment_check method correctly identifies spam.

        """
        check_kwargs = {
            # Akismet guarantees this will be classified spam.
            "comment_author": "viagra-test-123",
            **self.base_kwargs,
        }
        api = akismet.Akismet(http_client=_test_clients._make_test_sync_http_client())
        self.assertTrue(api.comment_check(**check_kwargs))

    def test_comment_check_not_spam(self):
        """
        The comment_check method correctly identifies non-spam.

        """
        check_kwargs = {
            # Akismet guarantees this will not be classified spam.
            "user_role": "administrator",
            **self.base_kwargs,
        }
        api = akismet.Akismet(
            http_client=_test_clients._make_test_sync_http_client(
                comment_check_response=_common.CheckResponse.HAM
            )
        )
        self.assertFalse(api.comment_check(**check_kwargs))

    def test_submit_spam(self):
        """
        The submit_spam method succeeds.

        """
        spam_kwargs = {
            "comment_type": "comment",
            "comment_author": "viagra-test-123",
            "comment_content": "viagra-test-123",
            **self.base_kwargs,
        }
        api = akismet.Akismet(http_client=_test_clients._make_test_sync_http_client())
        self.assertTrue(api.submit_spam(**spam_kwargs))

    def test_submit_ham(self):
        """
        The submit_ham method succeeds.

        """
        ham_kwargs = {
            "comment_type": "comment",
            "comment_author": "Legitimate Author",
            "comment_content": "This is a legitimate comment.",
            "user_role": "administrator",
            **self.base_kwargs,
        }
        api = akismet.Akismet(http_client=_test_clients._make_test_sync_http_client())
        self.assertTrue(api.submit_ham(**ham_kwargs))

    def test_unexpected_verify_key_response(self):
        """
        Unexpected verify_key API responses are correctly handled.

        """

        api = akismet.Akismet(
            http_client=_test_clients._make_test_sync_http_client(),
        )
        with self.assertRaises(akismet.ProtocolError):
            api.verify_key(
                self.api_key,
                self.site_url,
                http_client=make_fixed_response_sync_client(response_text="bad"),
            )

    def test_unexpected_comment_check_response(self):
        """
        Unexpected comment_check API responses are correctly handled.

        """
        api = akismet.Akismet(
            http_client=make_fixed_response_sync_client(response_text="valid"),
        )
        with self.assertRaises(akismet.ProtocolError):
            check_kwargs = {"comment_author": "viagra-test-123", **self.base_kwargs}
            api.comment_check(**check_kwargs)

    def test_unexpected_submit_spam_response(self):
        """
        Unexpected submit_spam API responses are correctly handled.

        """
        api = akismet.Akismet(
            http_client=make_fixed_response_sync_client(response_text="valid"),
        )
        with self.assertRaises(akismet.ProtocolError):
            spam_kwargs = {
                "comment_type": "comment",
                "comment_author": "viagra-test-123",
                "comment_content": "viagra-test-123",
                **self.base_kwargs,
            }
            api.submit_spam(**spam_kwargs)

    def test_unexpected_submit_ham_response(self):
        """
        Unexpected submit_ham API responses are correctly handled.

        """
        api = akismet.Akismet(
            http_client=make_fixed_response_sync_client(response_text="valid"),
        )
        with self.assertRaises(akismet.ProtocolError):
            ham_kwargs = {
                "comment_type": "comment",
                "comment_author": "Legitimate Author",
                "comment_content": "This is a legitimate comment.",
                "user_role": "administrator",
                **self.base_kwargs,
            }
            api.submit_ham(**ham_kwargs)

    def test_unknown_kwargs(self):
        """
        Unknown Akismet arguments are correctly rejected.

        """
        bad_kwargs = {"bad_arg": "bad_val", **self.base_kwargs}
        api = akismet.Akismet(
            http_client=_test_clients._make_test_sync_http_client(),
        )
        with self.assertRaises(akismet.UnknownArgumentError):
            api.comment_check(**bad_kwargs)
